@@ -26,6 +26,7 @@
 
 // kdeui
 #include "debug.h"
+#include "kiconcolors.h"
 #include "kiconeffect.h"
 #include "kicontheme.h"
 
@@ -50,6 +51,7 @@
 #include <QPixmap>
 #include <QPixmapCache>
 #include <QStringBuilder> // % operator for QString
+#include <QtGui/private/qiconloader_p.h>
 
 #include <qplatformdefs.h> //for readlink
 
@@ -63,32 +65,6 @@ QString NULL_EFFECT_FINGERPRINT()
     return QStringLiteral("noeffect");
 }
 
-QString STYLESHEET_TEMPLATE()
-{
-    /* clang-format off */
-    return QStringLiteral(".ColorScheme-Text {\
-color:%1;\
-}\
-.ColorScheme-Background{\
-color:%2;\
-}\
-.ColorScheme-Highlight{\
-color:%3;\
-}\
-.ColorScheme-HighlightedText{\
-color:%4;\
-}\
-.ColorScheme-PositiveText{\
-color:%5;\
-}\
-.ColorScheme-NeutralText{\
-color:%6;\
-}\
-.ColorScheme-NegativeText{\
-color:%7;\
-}");
-    /* clang-format on */
-}
 }
 
 /**
@@ -118,20 +94,20 @@ KICONTHEMES_EXPORT void uintToHex(uint32_t colorData, QChar *buffer)
     }
 }
 
-static QString paletteId(const QPalette &pal)
+static QString paletteId(const KIconColors &colors)
 {
     // 8 per color. We want 3 colors thus 8*4=32.
     QString buffer(32, Qt::Uninitialized);
 
-    uintToHex(pal.windowText().color().rgba(), buffer.data());
-    uintToHex(pal.highlight().color().rgba(), buffer.data() + 8);
-    uintToHex(pal.highlightedText().color().rgba(), buffer.data() + 16);
-    uintToHex(pal.window().color().rgba(), buffer.data() + 24);
+    uintToHex(colors.text().rgba(), buffer.data());
+    uintToHex(colors.highlight().rgba(), buffer.data() + 8);
+    uintToHex(colors.highlightedText().rgba(), buffer.data() + 16);
+    uintToHex(colors.background().rgba(), buffer.data() + 24);
 
     return buffer;
 }
 
-/*** KIconThemeNode: A node in the icon theme dependancy tree. ***/
+/*** KIconThemeNode: A node in the icon theme dependency tree. ***/
 
 class KIconThemeNode
 {
@@ -185,13 +161,99 @@ struct KIconGroup {
 extern KICONTHEMES_EXPORT int kiconloader_ms_between_checks;
 KICONTHEMES_EXPORT int kiconloader_ms_between_checks = 5000;
 
+class KIconLoaderGlobalData : public QObject
+{
+    Q_OBJECT
+
+public:
+    KIconLoaderGlobalData()
+    {
+        const QStringList genericIconsFiles = QStandardPaths::locateAll(QStandardPaths::GenericDataLocation, QStringLiteral("mime/generic-icons"));
+        // qCDebug(KICONTHEMES) << genericIconsFiles;
+        for (const QString &file : genericIconsFiles) {
+            parseGenericIconsFiles(file);
+        }
+
+#ifdef QT_DBUS_LIB
+        QDBusConnection::sessionBus().connect(QString(),
+                                              QStringLiteral("/KIconLoader"),
+                                              QStringLiteral("org.kde.KIconLoader"),
+                                              QStringLiteral("iconChanged"),
+                                              this,
+                                              SIGNAL(iconChanged(int)));
+#endif
+    }
+
+    void emitChange(KIconLoader::Group group)
+    {
+#ifdef QT_DBUS_LIB
+        QDBusMessage message = QDBusMessage::createSignal(QStringLiteral("/KIconLoader"), QStringLiteral("org.kde.KIconLoader"), QStringLiteral("iconChanged"));
+        message.setArguments(QList<QVariant>() << int(group));
+        QDBusConnection::sessionBus().send(message);
+#endif
+    }
+
+    QString genericIconFor(const QString &icon) const
+    {
+        return m_genericIcons.value(icon);
+    }
+
+Q_SIGNALS:
+    void iconChanged(int group);
+
+private:
+    void parseGenericIconsFiles(const QString &fileName);
+    QHash<QString, QString> m_genericIcons;
+};
+
+void KIconLoaderGlobalData::parseGenericIconsFiles(const QString &fileName)
+{
+    QFile file(fileName);
+    if (file.open(QIODevice::ReadOnly)) {
+        QTextStream stream(&file);
+        // In Qt6 the encoding is UTF-8 by default, so it should work for icon file names;
+        // I think this code had "ISO 8859-1" (i.e. Latin-1) as an optimization, but file
+        // names on Linux are UTF-8 by default, so this would be more robust.
+        // Note that in Qt6 we can have the same behaviour by using QTextStream::setEncoding().
+#if QT_VERSION < QT_VERSION_CHECK(6, 0, 0)
+        stream.setCodec("ISO 8859-1");
+#endif
+        while (!stream.atEnd()) {
+            const QString line = stream.readLine();
+            if (line.isEmpty() || line[0] == QLatin1Char('#')) {
+                continue;
+            }
+            const int pos = line.indexOf(QLatin1Char(':'));
+            if (pos == -1) { // syntax error
+                continue;
+            }
+            QString mimeIcon = line.left(pos);
+            const int slashindex = mimeIcon.indexOf(QLatin1Char('/'));
+            if (slashindex != -1) {
+                mimeIcon[slashindex] = QLatin1Char('-');
+            }
+
+            const QString genericIcon = line.mid(pos + 1);
+            m_genericIcons.insert(mimeIcon, genericIcon);
+            // qCDebug(KICONTHEMES) << mimeIcon << "->" << genericIcon;
+        }
+    }
+}
+
+Q_GLOBAL_STATIC(KIconLoaderGlobalData, s_globalData)
+
 /*** d pointer for KIconLoader. ***/
 class KIconLoaderPrivate
 {
 public:
-    KIconLoaderPrivate(KIconLoader *q)
-        : q(q)
+    KIconLoaderPrivate(const QString &_appname, const QStringList &extraSearchPaths, KIconLoader *qq)
+        : q(qq)
+        , m_appname(_appname)
     {
+        q->connect(s_globalData, &KIconLoaderGlobalData::iconChanged, q, [this](int group) {
+            _k_refreshIcons(group);
+        });
+        init(m_appname, extraSearchPaths);
     }
 
     ~KIconLoaderPrivate()
@@ -209,7 +271,7 @@ public:
         mpGroups = nullptr;
         mIconCache = nullptr;
         mPixmapCache.clear();
-        appname.clear();
+        m_appname.clear();
         searchPaths.clear();
         links.clear();
         mIconThemeInited = false;
@@ -300,7 +362,13 @@ public:
      * Used with KIconLoader::loadIcon to get a base key name from the given
      * icon metadata. Ensure the metadata is normalized first.
      */
-    QString makeCacheKey(const QString &name, KIconLoader::Group group, const QStringList &overlays, const QSize &size, qreal scale, int state) const;
+    QString makeCacheKey(const QString &name,
+                         KIconLoader::Group group,
+                         const QStringList &overlays,
+                         const QSize &size,
+                         qreal scale,
+                         int state,
+                         const KIconColors &colors) const;
 
     /**
      * @internal
@@ -309,7 +377,7 @@ public:
      * as text color, background color, highlight color, positive/neutral/negative color
      * @see KColorScheme
      */
-    QByteArray processSvg(const QString &path, KIconLoader::States state) const;
+    QByteArray processSvg(const QString &path, KIconLoader::States state, const KIconColors &colors) const;
 
     /**
      * @internal
@@ -317,7 +385,7 @@ public:
      * @p size is only used for scalable images, but if non-zero non-scalable
      * images will be resized anyways.
      */
-    QImage createIconImage(const QString &path, const QSize &size = {}, qreal scale = 1.0, KIconLoader::States state = KIconLoader::DefaultState);
+    QImage createIconImage(const QString &path, const QSize &size, qreal scale, KIconLoader::States state, const KIconColors &colors);
 
     /**
      * @internal
@@ -371,93 +439,19 @@ public:
     // lazy loading: initIconThemes() is only needed when the "links" list is needed
     // mIconThemeInited is used inside initIconThemes() to init only once
     bool mIconThemeInited : 1;
-    QString appname;
+    QString m_appname;
 
     void drawOverlays(const KIconLoader *loader, KIconLoader::Group group, int state, QPixmap &pix, const QStringList &overlays);
 
     QHash<QString, bool> mIconAvailability; // icon name -> true (known to be available) or false (known to be unavailable)
     QElapsedTimer mLastUnknownIconCheck; // recheck for unknown icons after kiconloader_ms_between_checks
-    // the palette used to recolor svg icons stylesheets
+    // the colors used to recolor svg icons stylesheets
+    KIconColors mColors;
     QPalette mPalette;
     // to keep track if we are using a custom palette or just falling back to qApp;
-    bool mCustomPalette = false;
+    bool mCustomColors = false;
 };
 
-class KIconLoaderGlobalData : public QObject
-{
-    Q_OBJECT
-
-public:
-    KIconLoaderGlobalData()
-    {
-        const QStringList genericIconsFiles = QStandardPaths::locateAll(QStandardPaths::GenericDataLocation, QStringLiteral("mime/generic-icons"));
-        // qCDebug(KICONTHEMES) << genericIconsFiles;
-        for (const QString &file : genericIconsFiles) {
-            parseGenericIconsFiles(file);
-        }
-
-#ifdef QT_DBUS_LIB
-        QDBusConnection::sessionBus().connect(QString(),
-                                              QStringLiteral("/KIconLoader"),
-                                              QStringLiteral("org.kde.KIconLoader"),
-                                              QStringLiteral("iconChanged"),
-                                              this,
-                                              SIGNAL(iconChanged(int)));
-#endif
-    }
-
-    void emitChange(KIconLoader::Group group)
-    {
-#ifdef QT_DBUS_LIB
-        QDBusMessage message = QDBusMessage::createSignal(QStringLiteral("/KIconLoader"), QStringLiteral("org.kde.KIconLoader"), QStringLiteral("iconChanged"));
-        message.setArguments(QList<QVariant>() << int(group));
-        QDBusConnection::sessionBus().send(message);
-#endif
-    }
-
-    QString genericIconFor(const QString &icon) const
-    {
-        return m_genericIcons.value(icon);
-    }
-
-Q_SIGNALS:
-    void iconChanged(int group);
-
-private:
-    void parseGenericIconsFiles(const QString &fileName);
-    QHash<QString, QString> m_genericIcons;
-};
-
-void KIconLoaderGlobalData::parseGenericIconsFiles(const QString &fileName)
-{
-    QFile file(fileName);
-    if (file.open(QIODevice::ReadOnly)) {
-        QTextStream stream(&file);
-        stream.setCodec("ISO 8859-1");
-        while (!stream.atEnd()) {
-            const QString line = stream.readLine();
-            if (line.isEmpty() || line[0] == QLatin1Char('#')) {
-                continue;
-            }
-            const int pos = line.indexOf(QLatin1Char(':'));
-            if (pos == -1) { // syntax error
-                continue;
-            }
-            QString mimeIcon = line.left(pos);
-            const int slashindex = mimeIcon.indexOf(QLatin1Char('/'));
-            if (slashindex != -1) {
-                mimeIcon[slashindex] = QLatin1Char('-');
-            }
-
-            const QString genericIcon = line.mid(pos + 1);
-            m_genericIcons.insert(mimeIcon, genericIcon);
-            // qCDebug(KICONTHEMES) << mimeIcon << "->" << genericIcon;
-        }
-    }
-}
-
-Q_GLOBAL_STATIC(KIconLoaderGlobalData, s_globalData)
-
 void KIconLoaderPrivate::drawOverlays(const KIconLoader *iconLoader, KIconLoader::Group group, int state, QPixmap &pix, const QStringList &overlays)
 {
     if (overlays.isEmpty()) {
@@ -543,10 +537,10 @@ void KIconLoaderPrivate::_k_refreshIcons(int group)
     sharedConfig->reparseConfiguration();
     const QString newThemeName = sharedConfig->group("Icons").readEntry("Theme", QStringLiteral("breeze"));
     if (!newThemeName.isEmpty()) {
-        // If we're refreshing icons the Qt platform plugin has probably
-        // already cached the old theme, which will accidentally filter back
-        // into KIconTheme unless we reset it
-        QIcon::setThemeName(newThemeName);
+        // NOTE Do NOT use QIcon::setThemeName here it makes Qt not use icon engine of the platform theme
+        //      anymore (KIconEngine on Plasma, which breaks recoloring) and overwrites a user set themeName
+        // TODO KF6 this should be done in the Plasma QPT
+        QIconLoader::instance()->updateSystemTheme();
     }
 
     q->newIconLoader();
@@ -554,14 +548,11 @@ void KIconLoaderPrivate::_k_refreshIcons(int group)
     Q_EMIT q->iconChanged(group);
 }
 
-KIconLoader::KIconLoader(const QString &_appname, const QStringList &extraSearchPaths, QObject *parent)
+KIconLoader::KIconLoader(const QString &appname, const QStringList &extraSearchPaths, QObject *parent)
     : QObject(parent)
-    , d(new KIconLoaderPrivate(this))
+    , d(new KIconLoaderPrivate(appname, extraSearchPaths, this))
 {
-    setObjectName(_appname);
-
-    connect(s_globalData, SIGNAL(iconChanged(int)), SLOT(_k_refreshIcons(int)));
-    d->init(_appname, extraSearchPaths);
+    setObjectName(appname);
 }
 
 void KIconLoader::reconfigure(const QString &_appname, const QStringList &extraSearchPaths)
@@ -579,10 +570,7 @@ void KIconLoaderPrivate::init(const QString &_appname, const QStringList &extraS
 
     searchPaths = extraSearchPaths;
 
-    appname = _appname;
-    if (appname.isEmpty()) {
-        appname = QCoreApplication::applicationName();
-    }
+    m_appname = !_appname.isEmpty() ? _appname : QCoreApplication::applicationName();
 
     // Initialize icon cache
     mIconCache = new KSharedDataCache(QStringLiteral("icon-cache"), 10 * 1024 * 1024);
@@ -621,12 +609,12 @@ void KIconLoaderPrivate::initIconThemes()
     mIconThemeInited = true;
 
     // Add the default theme and its base themes to the theme tree
-    KIconTheme *def = new KIconTheme(KIconTheme::current(), appname);
+    KIconTheme *def = new KIconTheme(KIconTheme::current(), m_appname);
     if (!def->isValid()) {
         delete def;
         // warn, as this is actually a small penalty hit
         qCDebug(KICONTHEMES) << "Couldn't find current icon theme, falling back to default.";
-        def = new KIconTheme(KIconTheme::defaultThemeName(), appname);
+        def = new KIconTheme(KIconTheme::defaultThemeName(), m_appname);
         if (!def->isValid()) {
             qCDebug(KICONTHEMES) << "Standard icon theme" << KIconTheme::defaultThemeName() << "not found!";
             delete def;
@@ -636,10 +624,10 @@ void KIconLoaderPrivate::initIconThemes()
     mpThemeRoot = new KIconThemeNode(def);
     mThemesInTree.append(def->internalName());
     links.append(mpThemeRoot);
-    addBaseThemes(mpThemeRoot, appname);
+    addBaseThemes(mpThemeRoot, m_appname);
 
     // Insert application specific themes at the top.
-    searchPaths.append(appname + QStringLiteral("/pics"));
+    searchPaths.append(m_appname + QStringLiteral("/pics"));
 
     // Add legacy icon dirs.
     searchPaths.append(QStringLiteral("icons")); // was xdgdata-icon in KStandardDirs
@@ -656,16 +644,12 @@ QStringList KIconLoader::searchPaths() const
 
 void KIconLoader::addAppDir(const QString &appname, const QString &themeBaseDir)
 {
-    d->initIconThemes();
-
     d->searchPaths.append(appname + QStringLiteral("/pics"));
     d->addAppThemes(appname, themeBaseDir);
 }
 
 void KIconLoaderPrivate::addAppThemes(const QString &appname, const QString &themeBaseDir)
 {
-    initIconThemes();
-
     KIconTheme *def = new KIconTheme(QStringLiteral("hicolor"), appname, themeBaseDir);
     if (!def->isValid()) {
         delete def;
@@ -700,21 +684,22 @@ void KIconLoaderPrivate::addBaseThemes(KIconThemeNode *node, const QString &appn
     // present in an inherited theme, and hicolor goes last.
 
     addInheritedThemes(node, appname);
+    addThemeByName(QIcon::fallbackThemeName(), appname);
     addThemeByName(QStringLiteral("hicolor"), appname);
 }
 
 void KIconLoaderPrivate::addInheritedThemes(KIconThemeNode *node, const QString &appname)
 {
-    const QStringList lst = node->theme->inherits();
+    const QStringList inheritedThemes = node->theme->inherits();
 
-    for (QStringList::ConstIterator it = lst.begin(), total = lst.end(); it != total; ++it) {
-        if ((*it) == QLatin1String("hicolor")) {
+    for (const auto &inheritedTheme : inheritedThemes) {
+        if (inheritedTheme == QLatin1String("hicolor")) {
             // The icon theme spec says that "hicolor" must be the very last
             // of all inherited themes, so don't add it here but at the very end
             // of addBaseThemes().
             continue;
         }
-        addThemeByName(*it, appname);
+        addThemeByName(inheritedTheme, appname);
     }
 }
 
@@ -740,44 +725,34 @@ void KIconLoaderPrivate::addExtraDesktopThemes()
         return;
     }
 
-    initIconThemes();
-
     QStringList list;
     const QStringList icnlibs = QStandardPaths::locateAll(QStandardPaths::GenericDataLocation, QStringLiteral("icons"), QStandardPaths::LocateDirectory);
-    char buf[1000];
-    for (QStringList::ConstIterator it = icnlibs.begin(), total = icnlibs.end(); it != total; ++it) {
-        QDir dir(*it);
+    for (const auto &iconDir : icnlibs) {
+        QDir dir(iconDir);
         if (!dir.exists()) {
             continue;
         }
-        const QStringList lst = dir.entryList(QStringList(QStringLiteral("default.*")), QDir::Dirs);
-        for (QStringList::ConstIterator it2 = lst.begin(), total = lst.end(); it2 != total; ++it2) {
-            if (!QFileInfo::exists(*it + *it2 + QStringLiteral("/index.desktop")) //
-                && !QFileInfo::exists(*it + *it2 + QStringLiteral("/index.theme"))) {
+        const auto defaultEntries = dir.entryInfoList(QStringList(QStringLiteral("default.*")), QDir::Dirs);
+        for (const auto &defaultEntry : defaultEntries) {
+            if (!QFileInfo::exists(defaultEntry.filePath() + QLatin1String("/index.desktop")) //
+                && !QFileInfo::exists(defaultEntry.filePath() + QLatin1String("/index.theme"))) {
                 continue;
             }
-            // TODO: Is any special handling required for NTFS symlinks?
-#ifndef Q_OS_WIN
-            const int r = readlink(QFile::encodeName(*it + *it2), buf, sizeof(buf) - 1);
-            if (r > 0) {
-                buf[r] = 0;
-                const QDir dir2(buf);
-                QString themeName = dir2.dirName();
-
+            if (defaultEntry.isSymbolicLink()) {
+                const QString themeName = QDir(defaultEntry.symLinkTarget()).dirName();
                 if (!list.contains(themeName)) {
                     list.append(themeName);
                 }
             }
-#endif
         }
     }
 
-    for (QStringList::ConstIterator it = list.constBegin(), total = list.constEnd(); it != total; ++it) {
+    for (const auto &theme : list) {
         // Don't add the KDE defaults once more, we have them anyways.
-        if (*it == QLatin1String("default.kde") || *it == QLatin1String("default.kde4")) {
+        if (theme == QLatin1String("default.kde") || theme == QLatin1String("default.kde4")) {
             continue;
         }
-        addThemeByName(*it, QLatin1String(""));
+        addThemeByName(theme, QLatin1String(""));
     }
 
     extraDesktopIconsLoaded = true;
@@ -804,7 +779,7 @@ QString KIconLoaderPrivate::removeIconExtension(const QString &name) const
 void KIconLoaderPrivate::normalizeIconMetadata(KIconLoader::Group &group, QSize &size, int &state) const
 {
     if ((state < 0) || (state >= KIconLoader::LastState)) {
-        qWarning() << "Illegal icon state:" << state;
+        qCWarning(KICONTHEMES) << "Invalid icon state:" << state << ", should be one of KIconLoader::States";
         state = KIconLoader::DefaultState;
     }
 
@@ -819,7 +794,7 @@ void KIconLoaderPrivate::normalizeIconMetadata(KIconLoader::Group &group, QSize
     }
 
     if ((group < -1) || (group >= KIconLoader::LastGroup)) {
-        qWarning() << "Illegal icon group:" << group;
+        qCWarning(KICONTHEMES) << "Invalid icon group:" << group << ", should be one of KIconLoader::Group";
         group = KIconLoader::Desktop;
     }
 
@@ -833,7 +808,13 @@ void KIconLoaderPrivate::normalizeIconMetadata(KIconLoader::Group &group, QSize
     }
 }
 
-QString KIconLoaderPrivate::makeCacheKey(const QString &name, KIconLoader::Group group, const QStringList &overlays, const QSize &size, qreal scale, int state) const
+QString KIconLoaderPrivate::makeCacheKey(const QString &name,
+                                         KIconLoader::Group group,
+                                         const QStringList &overlays,
+                                         const QSize &size,
+                                         qreal scale,
+                                         int state,
+                                         const KIconColors &colors) const
 {
     // The KSharedDataCache is shared so add some namespacing. The following code
     // uses QStringBuilder (new in Qt 4.6)
@@ -849,14 +830,14 @@ QString KIconLoaderPrivate::makeCacheKey(const QString &name, KIconLoader::Group
             % overlays.join(QLatin1Char('_'))
             % (group >= 0 ? mpEffect.fingerprint(group, state) : NULL_EFFECT_FINGERPRINT())
             % QLatin1Char('_')
-            % paletteId(mCustomPalette ? mPalette : qApp->palette())
+            % paletteId(colors)
             % (q->theme() && q->theme()->followsColorScheme() && state == KIconLoader::SelectedState ? QStringLiteral("_selected") : QString());
     /* clang-format on */
 }
 
-QByteArray KIconLoaderPrivate::processSvg(const QString &path, KIconLoader::States state) const
+QByteArray KIconLoaderPrivate::processSvg(const QString &path, KIconLoader::States state, const KIconColors &colors) const
 {
-    QScopedPointer<QIODevice> device;
+    std::unique_ptr<QIODevice> device;
 
     if (path.endsWith(QLatin1String("svgz"))) {
         device.reset(new KCompressionDevice(path, KCompressionDevice::GZip));
@@ -868,22 +849,9 @@ QByteArray KIconLoaderPrivate::processSvg(const QString &path, KIconLoader::Stat
         return QByteArray();
     }
 
-    const QPalette pal = mCustomPalette ? mPalette : qApp->palette();
-    KColorScheme scheme(QPalette::Active, KColorScheme::Window);
-
-    /* clang-format off */
-    QString styleSheet = STYLESHEET_TEMPLATE().arg(
-                                state == KIconLoader::SelectedState ? pal.highlightedText().color().name() : pal.windowText().color().name(),
-                                state == KIconLoader::SelectedState ? pal.highlight().color().name() : pal.window().color().name(),
-                                state == KIconLoader::SelectedState ? pal.highlightedText().color().name() : pal.highlight().color().name(),
-                                state == KIconLoader::SelectedState ? pal.highlight().color().name() : pal.highlightedText().color().name(),
-                                scheme.foreground(KColorScheme::PositiveText).color().name(),
-                                scheme.foreground(KColorScheme::NeutralText).color().name(),
-                                scheme.foreground(KColorScheme::NegativeText).color().name());
-    /* clang-format on */
-
+    const QString styleSheet = colors.stylesheet(state);
     QByteArray processedContents;
-    QXmlStreamReader reader(device.data());
+    QXmlStreamReader reader(device.get());
 
     QBuffer buffer(&processedContents);
     buffer.open(QIODevice::WriteOnly);
@@ -908,14 +876,14 @@ QByteArray KIconLoaderPrivate::processSvg(const QString &path, KIconLoader::Stat
     return processedContents;
 }
 
-QImage KIconLoaderPrivate::createIconImage(const QString &path, const QSize &size, qreal scale, KIconLoader::States state)
+QImage KIconLoaderPrivate::createIconImage(const QString &path, const QSize &size, qreal scale, KIconLoader::States state, const KIconColors &colors)
 {
     // TODO: metadata in the theme to make it do this only if explicitly supported?
     QImageReader reader;
     QBuffer buffer;
 
     if (q->theme() && q->theme()->followsColorScheme() && (path.endsWith(QLatin1String("svg")) || path.endsWith(QLatin1String("svgz")))) {
-        buffer.setData(processSvg(path, state));
+        buffer.setData(processSvg(path, state, colors));
         reader.setDevice(&buffer);
     } else {
         reader.setFileName(path);
@@ -1041,39 +1009,27 @@ QString KIconLoaderPrivate::findMatchingIconWithGenericFallbacks(const QString &
 
 QString KIconLoaderPrivate::findMatchingIcon(const QString &name, int size, qreal scale) const
 {
-    const_cast<KIconLoaderPrivate *>(this)->initIconThemes();
-
-    // Do two passes through themeNodes.
-    //
-    // The first pass looks for an exact match in each themeNode one after the other.
-    // If one is found and it is an app icon then return that icon.
-    //
-    // In the next pass (assuming the first pass failed), it looks for
+    // This looks for the exact match and its
     // generic fallbacks in each themeNode one after the other.
 
     // In theory we should only do this for mimetype icons, not for app icons,
     // but that would require different APIs. The long term solution is under
     // development for Qt >= 5.8, QFileIconProvider calling QPlatformTheme::fileIcon,
     // using QMimeType::genericIconName() to get the proper -x-generic fallback.
-    // Once everone uses that to look up mimetype icons, we can kill the fallback code
+    // Once everyone uses that to look up mimetype icons, we can kill the fallback code
     // from this method.
 
-    for (KIconThemeNode *themeNode : qAsConst(links)) {
-        const QString path = themeNode->theme->iconPathByName(name, size, KIconLoader::MatchBest, scale);
-        if (!path.isEmpty()) {
-            return path;
-        }
-    }
-
-    if (name.endsWith(QLatin1String("-x-generic"))) {
-        return QString(); // no further fallback
-    }
-    bool genericFallback = false;
+    bool genericFallback = name.endsWith(QLatin1String("-x-generic"));;
     QString path;
-    for (KIconThemeNode *themeNode : qAsConst(links)) {
+    for (KIconThemeNode *themeNode : std::as_const(links)) {
         QString currentName = name;
 
         while (!currentName.isEmpty()) {
+            path = themeNode->theme->iconPathByName(currentName, size, KIconLoader::MatchBest, scale);
+            if (!path.isEmpty()) {
+                return path;
+            }
+
             if (genericFallback) {
                 // we already tested the base name
                 break;
@@ -1110,16 +1066,6 @@ QString KIconLoaderPrivate::findMatchingIcon(const QString &name, int size, qrea
                     break;
                 }
             }
-
-            if (currentName.isEmpty()) {
-                break;
-            }
-
-            // qCDebug(KICONTHEMES) << "Looking up" << currentName;
-            path = themeNode->theme->iconPathByName(currentName, size, KIconLoader::MatchBest, scale);
-            if (!path.isEmpty()) {
-                return path;
-            }
         }
     }
 
@@ -1154,7 +1100,7 @@ inline QString KIconLoaderPrivate::unknownIconPath(int size, qreal scale) const
 
 QString KIconLoaderPrivate::locate(const QString &fileName)
 {
-    for (const QString &dir : qAsConst(searchPaths)) {
+    for (const QString &dir : std::as_const(searchPaths)) {
         const QString path = dir + QLatin1Char('/') + fileName;
         if (QDir(dir).isAbsolute()) {
             if (QFileInfo::exists(path)) {
@@ -1185,8 +1131,6 @@ QString KIconLoader::iconPath(const QString &_name, int group_or_size, bool canR
         return _name;
     }
 
-    d->initIconThemes();
-
     QString name = d->removeIconExtension(_name);
 
     QString path;
@@ -1205,7 +1149,7 @@ QString KIconLoader::iconPath(const QString &_name, int group_or_size, bool canR
     }
 
     if (group_or_size >= KIconLoader::LastGroup) {
-        qCDebug(KICONTHEMES) << "Illegal icon group:" << group_or_size;
+        qCDebug(KICONTHEMES) << "Invalid icon group:" << group_or_size;
         return path;
     }
 
@@ -1288,11 +1232,24 @@ QPixmap KIconLoader::loadScaledIcon(const QString &_name,
 QPixmap KIconLoader::loadScaledIcon(const QString &_name,
                                     KIconLoader::Group group,
                                     qreal scale,
-                                    const QSize &_size,
+                                    const QSize &size,
                                     int state,
                                     const QStringList &overlays,
                                     QString *path_store,
                                     bool canReturnNull) const
+{
+    return loadScaledIcon(_name, group, scale, size, state, overlays, path_store, canReturnNull, {});
+}
+
+QPixmap KIconLoader::loadScaledIcon(const QString &_name,
+                                    KIconLoader::Group group,
+                                    qreal scale,
+                                    const QSize &_size,
+                                    int state,
+                                    const QStringList &overlays,
+                                    QString *path_store,
+                                    bool canReturnNull,
+                                    const std::optional<KIconColors> &colors) const
 
 {
     QString name = _name;
@@ -1336,7 +1293,8 @@ QPixmap KIconLoader::loadScaledIcon(const QString &_name,
     d->normalizeIconMetadata(group, size, state);
 
     // See if the image is already cached.
-    QString key = d->makeCacheKey(name, group, overlays, size, scale, state);
+    auto usedColors = colors ? *colors : d->mCustomColors ? d->mColors : KIconColors(qApp->palette());
+    QString key = d->makeCacheKey(name, group, overlays, size, scale, state, usedColors);
     QPixmap pix;
 
     bool iconWasUnknown = false;
@@ -1361,7 +1319,6 @@ QPixmap KIconLoader::loadScaledIcon(const QString &_name,
     }
 
     // Image is not cached... go find it and apply effects.
-    d->initIconThemes();
 
     favIconOverlay = favIconOverlay && std::min(size.height(), size.width()) > 22;
 
@@ -1389,7 +1346,7 @@ QPixmap KIconLoader::loadScaledIcon(const QString &_name,
 
     QImage img;
     if (!path.isEmpty()) {
-        img = d->createIconImage(path, size, scale, static_cast<KIconLoader::States>(state));
+        img = d->createIconImage(path, size, scale, static_cast<KIconLoader::States>(state), usedColors);
     }
 
     if (group >= 0 && group < KIconLoader::LastGroup) {
@@ -1463,10 +1420,8 @@ QString KIconLoader::moviePath(const QString &name, KIconLoader::Group group, in
         return QString();
     }
 
-    d->initIconThemes();
-
     if ((group < -1 || group >= KIconLoader::LastGroup) && group != KIconLoader::User) {
-        qCDebug(KICONTHEMES) << "Illegal icon group:" << group;
+        qCDebug(KICONTHEMES) << "Invalid icon group:" << group << ", should be one of KIconLoader::Group";
         group = KIconLoader::Desktop;
     }
     if (size == 0 && group < 0) {
@@ -1484,7 +1439,7 @@ QString KIconLoader::moviePath(const QString &name, KIconLoader::Group group, in
 
         QString path;
 
-        for (KIconThemeNode *themeNode : qAsConst(d->links)) {
+        for (KIconThemeNode *themeNode : std::as_const(d->links)) {
             path = themeNode->theme->iconPath(file, size, KIconLoader::MatchExact);
             if (!path.isEmpty()) {
                 break;
@@ -1492,7 +1447,7 @@ QString KIconLoader::moviePath(const QString &name, KIconLoader::Group group, in
         }
 
         if (path.isEmpty()) {
-            for (KIconThemeNode *themeNode : qAsConst(d->links)) {
+            for (KIconThemeNode *themeNode : std::as_const(d->links)) {
                 path = themeNode->theme->iconPath(file, size, KIconLoader::MatchBest);
                 if (!path.isEmpty()) {
                     break;
@@ -1516,7 +1471,7 @@ QStringList KIconLoader::loadAnimated(const QString &name, KIconLoader::Group gr
     d->initIconThemes();
 
     if ((group < -1) || (group >= KIconLoader::LastGroup)) {
-        qCDebug(KICONTHEMES) << "Illegal icon group: " << group;
+        qCDebug(KICONTHEMES) << "Invalid icon group: " << group << ", should be one of KIconLoader::Group";
         group = KIconLoader::Desktop;
     }
     if ((size == 0) && (group < 0)) {
@@ -1545,7 +1500,12 @@ QStringList KIconLoader::loadAnimated(const QString &name, KIconLoader::Group gr
 
     const auto entryList = dir.entryList();
     for (const QString &entry : entryList) {
-        if (!(entry.leftRef(4)).toUInt()) {
+#if QT_VERSION >= QT_VERSION_CHECK(6, 0, 0)
+        const QStringView chunk = QStringView(entry).left(4);
+#else
+        const QStringRef chunk = entry.leftRef(4);
+#endif
+        if (!chunk.toUInt()) {
             continue;
         }
 
@@ -1557,7 +1517,6 @@ QStringList KIconLoader::loadAnimated(const QString &name, KIconLoader::Group gr
 
 KIconTheme *KIconLoader::theme() const
 {
-    d->initIconThemes();
     if (d->mpThemeRoot) {
         return d->mpThemeRoot->theme;
     }
@@ -1571,7 +1530,7 @@ int KIconLoader::currentSize(KIconLoader::Group group) const
     }
 
     if (group < 0 || group >= KIconLoader::LastGroup) {
-        qCDebug(KICONTHEMES) << "Illegal icon group:" << group;
+        qCDebug(KICONTHEMES) << "Invalid icon group:" << group << ", should be one of KIconLoader::Group";
         return -1;
     }
     return d->mpGroups[group].size;
@@ -1583,19 +1542,17 @@ QStringList KIconLoader::queryIconsByDir(const QString &iconsDir) const
     const QStringList formats = QStringList() << QStringLiteral("*.png") << QStringLiteral("*.xpm") << QStringLiteral("*.svg") << QStringLiteral("*.svgz");
     const QStringList lst = dir.entryList(formats, QDir::Files);
     QStringList result;
-    for (QStringList::ConstIterator it = lst.begin(), total = lst.end(); it != total; ++it) {
-        result += iconsDir + QLatin1Char('/') + *it;
+    for (const auto &file : lst) {
+        result += iconsDir + QLatin1Char('/') + file;
     }
     return result;
 }
 
 QStringList KIconLoader::queryIconsByContext(int group_or_size, KIconLoader::Context context) const
 {
-    d->initIconThemes();
-
     QStringList result;
     if (group_or_size >= KIconLoader::LastGroup) {
-        qCDebug(KICONTHEMES) << "Illegal icon group:" << group_or_size;
+        qCDebug(KICONTHEMES) << "Invalid icon group:" << group_or_size;
         return result;
     }
     int size;
@@ -1605,25 +1562,25 @@ QStringList KIconLoader::queryIconsByContext(int group_or_size, KIconLoader::Con
         size = -group_or_size;
     }
 
-    for (KIconThemeNode *themeNode : qAsConst(d->links)) {
+    for (KIconThemeNode *themeNode : std::as_const(d->links)) {
         themeNode->queryIconsByContext(&result, size, context);
     }
 
     // Eliminate duplicate entries (same icon in different directories)
     QString name;
-    QStringList res2, entries;
-    QStringList::ConstIterator it;
-    for (it = result.constBegin(); it != result.constEnd(); ++it) {
-        int n = (*it).lastIndexOf(QLatin1Char('/'));
+    QStringList res2;
+    QStringList entries;
+    for (const auto &icon : std::as_const(result)) {
+        const int n = icon.lastIndexOf(QLatin1Char('/'));
         if (n == -1) {
-            name = *it;
+            name = icon;
         } else {
-            name = (*it).mid(n + 1);
+            name = icon.mid(n + 1);
         }
         name = d->removeIconExtension(name);
         if (!entries.contains(name)) {
             entries += name;
-            res2 += *it;
+            res2 += icon;
         }
     }
     return res2;
@@ -1635,7 +1592,7 @@ QStringList KIconLoader::queryIcons(int group_or_size, KIconLoader::Context cont
 
     QStringList result;
     if (group_or_size >= KIconLoader::LastGroup) {
-        qCDebug(KICONTHEMES) << "Illegal icon group:" << group_or_size;
+        qCDebug(KICONTHEMES) << "Invalid icon group:" << group_or_size;
         return result;
     }
     int size;
@@ -1645,25 +1602,25 @@ QStringList KIconLoader::queryIcons(int group_or_size, KIconLoader::Context cont
         size = -group_or_size;
     }
 
-    for (KIconThemeNode *themeNode : qAsConst(d->links)) {
+    for (KIconThemeNode *themeNode : std::as_const(d->links)) {
         themeNode->queryIcons(&result, size, context);
     }
 
     // Eliminate duplicate entries (same icon in different directories)
     QString name;
-    QStringList res2, entries;
-    QStringList::ConstIterator it;
-    for (it = result.constBegin(); it != result.constEnd(); ++it) {
-        int n = (*it).lastIndexOf(QLatin1Char('/'));
+    QStringList res2;
+    QStringList entries;
+    for (const auto &icon : std::as_const(result)) {
+        const int n = icon.lastIndexOf(QLatin1Char('/'));
         if (n == -1) {
-            name = *it;
+            name = icon;
         } else {
-            name = (*it).mid(n + 1);
+            name = icon.mid(n + 1);
         }
         name = d->removeIconExtension(name);
         if (!entries.contains(name)) {
             entries += name;
-            res2 += *it;
+            res2 += icon;
         }
     }
     return res2;
@@ -1672,12 +1629,11 @@ QStringList KIconLoader::queryIcons(int group_or_size, KIconLoader::Context cont
 // used by KIconDialog to find out which contexts to offer in a combobox
 bool KIconLoader::hasContext(KIconLoader::Context context) const
 {
-    d->initIconThemes();
-
-    for (KIconThemeNode *themeNode : qAsConst(d->links))
+    for (KIconThemeNode *themeNode : std::as_const(d->links)) {
         if (themeNode->theme->hasContext(context)) {
             return true;
         }
+    }
     return false;
 }
 
@@ -1686,6 +1642,7 @@ KIconEffect *KIconLoader::iconEffect() const
     return &d->mpEffect;
 }
 
+#if KICONTHEMES_BUILD_DEPRECATED_SINCE(5, 82)
 bool KIconLoader::alphaBlending(KIconLoader::Group group) const
 {
     if (!d->mpGroups) {
@@ -1693,11 +1650,12 @@ bool KIconLoader::alphaBlending(KIconLoader::Group group) const
     }
 
     if (group < 0 || group >= KIconLoader::LastGroup) {
-        qCDebug(KICONTHEMES) << "Illegal icon group:" << group;
+        qCDebug(KICONTHEMES) << "Invalid icon group:" << group << ", should be one of KIconLoader::Group";
         return false;
     }
     return true;
 }
+#endif
 
 #if KICONTHEMES_BUILD_DEPRECATED_SINCE(5, 0)
 QIcon KIconLoader::loadIconSet(const QString &name, KIconLoader::Group g, int s, bool canReturnNull)
@@ -1842,18 +1800,23 @@ bool KIconLoader::hasIcon(const QString &name) const
 
 void KIconLoader::setCustomPalette(const QPalette &palette)
 {
-    d->mCustomPalette = true;
-    d->mPalette = palette;
+    d->mCustomColors = true;
+    d->mColors = KIconColors(palette);
 }
 
 QPalette KIconLoader::customPalette() const
 {
-    return d->mCustomPalette ? d->mPalette : QPalette();
+    return d->mCustomColors ? d->mPalette : QPalette();
 }
 
 void KIconLoader::resetPalette()
 {
-    d->mCustomPalette = false;
+    d->mCustomColors = false;
+}
+
+bool KIconLoader::hasCustomPalette() const
+{
+    return d->mCustomColors;
 }
 
 /*** the global icon loader ***/
@@ -1890,5 +1853,10 @@ QIcon KDE::icon(const QString &iconName, const QStringList &overlays, KIconLoade
     return QIcon(new KIconEngine(iconName, iconLoader ? iconLoader : KIconLoader::global(), overlays));
 }
 
+QIcon KDE::icon(const QString &iconName, const KIconColors &colors, KIconLoader *iconLoader)
+{
+    return QIcon(new KIconEngine(iconName, colors, iconLoader ? iconLoader : KIconLoader::global()));
+}
+
 #include "kiconloader.moc"
 #include "moc_kiconloader.moc"
